<?php
/**
 * @brief		Background Task: Rebuild Attachment Thumbnails
 * @author		<a href='http://www.invisionpower.com'>Invision Power Services, Inc.</a>
 * @copyright	(c) 2001 - 2016 Invision Power Services, Inc.
 * @license		http://www.invisionpower.com/legal/standards/
 * @package		IPS Community Suite
 * @since		24 Nov 2015
 * @version		SVN_VERSION_NUMBER
 */

namespace IPS\core\extensions\core\Queue;

/* To prevent PHP errors (extending class does not exist) revealing path */
if ( !defined( '\IPS\SUITE_UNIQUE_KEY' ) )
{
	header( ( isset( $_SERVER['SERVER_PROTOCOL'] ) ? $_SERVER['SERVER_PROTOCOL'] : 'HTTP/1.0' ) . ' 403 Forbidden' );
	exit;
}

/**
 * Background Task: Rebuilds thumbnails for attached images
 */
class _RebuildAttachmentThumbnails
{
	/**
	 * @brief Number of thumbnails to build per cycle
	 */
	public $perCycle	= 5;
	
	/**
	 * Parse data before queuing
	 *
	 * @param	array	$data
	 * @return	array
	 */
	public function preQueueData( $data )
	{
		try
		{			
			$data['count']		= \IPS\Db::i()->select( 'COUNT(*)', 'core_attachments', array( 'attach_is_image=?', 1 ) )->first();
		}
		catch( \Exception $ex )
		{
			throw new \OutOfRangeException;
		}
		
		if( $data['count'] == 0 )
		{
			return null;
		}

		return $data;
	}

	/**
	 * Run Background Task
	 *
	 * @param	mixed					$data	Data as it was passed to \IPS\Task::queue()
	 * @param	int						$offset	Offset
	 * @return	int|null				New offset or NULL if complete
	 * @throws	\OutOfRangeException	Indicates offset doesn't exist and thus task is complete
	 */
	public function run( &$data, $offset )
	{
		$count = 0;
		
		\IPS\Log::i( LOG_DEBUG )->write( "Rebuilding attachment thumbnails with an offset of " . $offset, 'rebuildAttachmentThumbnails' );

		$thumbDims	= \IPS\Settings::i()->attachment_image_size ? explode( 'x', \IPS\Settings::i()->attachment_image_size ) : array( 1000, 750 );

		foreach( \IPS\Db::i()->select( '*', 'core_attachments', array( 'attach_is_image=?', 1 ), 'attach_id ASC', array( $offset, $this->perCycle ) ) as $attachment )
		{
			$count++;

			try
			{
				$file = \IPS\File::get( 'core_Attachment', $attachment['attach_location'] );

				/* If there is no existing thumb location AND the existing dimensions are smaller than those needed for a thumbnail,
					we don't need to build a thumbnail */
				if( !$attachment['attach_thumb_location'] )
				{
					$thumbDims	= \IPS\Settings::i()->attachment_image_size ? explode( 'x', \IPS\Settings::i()->attachment_image_size ) : array( 1000, 750 );
					$dimensions	= $file->getImageDimensions();

					if ( $dimensions[0] < $thumbDims[0] and $dimensions[1] < $thumbDims[1] )
					{
						continue;
					}
				}

				if( $attachment['attach_thumb_location'] )
				{
					$thumb = \IPS\File::get( 'core_Attachment', $attachment['attach_thumb_location'] );
					$file->thumbnailName		= $thumb->filename;
					$file->thumbnailContainer	= $thumb->container;
				}

				\IPS\Db::i()->update( 'core_attachments', array(
					'attach_thumb_location'	=> (string) $file->thumbnail( 'core_Attachment', $thumbDims[0], $thumbDims[1] ),
				), array( "attach_id=?", $attachment['attach_id'] ) );
			}
			catch( \RuntimeException $e )
			{
				continue;
			}
			catch( \InvalidArgumentException $e )
			{
				continue;
			}
		}

		return $count ? ( $offset + $count ) : NULL;
	}
	
	/**
	 * Get Progress
	 *
	 * @param	mixed					$data	Data as it was passed to \IPS\Task::queue()
	 * @param	int						$offset	Offset
	 * @return	array( 'text' => 'Doing something...', 'complete' => 50 )	Text explaining task and percentage complete
	 * @throws	\OutOfRangeException	Indicates offset doesn't exist and thus task is complete
	 */
	public function getProgress( $data, $offset )
	{
		return array( 'text' => \IPS\Member::loggedIn()->language()->addToStack('rebuilding_attachments'), 'complete' => round( 100 / $data['count'] * $offset, 2 ) );
	}	
}